/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.vew.bmvew.bmvew.sender;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TimerTask;
import java.util.TreeSet;
import java.util.Vector;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.Data.TextValue;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.dambach.vewdynobj.VerwaltungDynObj;
import de.bsvrz.vew.bmvew.bmvew.BetriebsMeldungApp;
import de.bsvrz.vew.bmvew.bmvew.meldung.Meldung;


/**
 * Die Klasse zum zeitgesteuerten Versenden von Datenstzen. 
 * Erweitert die Klasse TimerTask. Die run() Methode wird  ausgefuehrt wenn die im 
 * der Klasse ErinnerungsFunktion eingestellte Zeit abgelaufen ist. Sie implentiert 
 * zustzlich das  ClientSenderInterface welches die Methoden zum Versenden der Daten 
 * bereitstellt. 
 * 
 * @version  $Revision: 1.3 $ / $Date: 2013/05/23 13:47:44 $ / ($Author: Pittner $)
 * @author  Dambach-Werke GmbH
 * @author  Timo Pittner
 */
public class SenderZeitGesteuert extends TimerTask implements ClientSenderInterface, MutableSetChangeListener
{
  /**
   * Die Daten EreignisKalenderAnfrage
   */
  private ClientDavInterface _connection;
  
  /**
   * Die Daten EreignisKalenderAnfrage
   */
  private DataModel _datenmodell;
  
  
  /**
   * Die Meldung
   */
  private Meldung _meldung;
  
  
  /**
   * Die Daten Meldung
   */
  private ResultData _resultData;


  /**
   * Die Daten EreignisKalenderAnfrage
   */
  private DataDescription _dd;

  
  /**
   * Der Debugger, stellt verschiedene Protokollierungsfunktionen zur Verfgung
   */
  private Debug _debug;
  
  /**
   * Die Daten EreignisKalenderAnfrage
   */
  private boolean _geloescht = false;

    
  /**
   * Konstruktor der Klasse
   * 
   * @param result
   *          Die Daten EreignisKalenderAnfrage
   */
  public SenderZeitGesteuert(ClientDavInterface connection, Meldung meldung, DataDescription dd)
  {    
    _connection = connection;
    _datenmodell = _connection.getDataModel();
    _meldung = meldung;
    _resultData = meldung.getResult();
    _dd = dd;
    _debug = Debug.getLogger();
    
    MutableSet ms = BetriebsMeldungApp.getKonfigurationsObjekt().getMutableSet("Meldungen");

    if (ms != null)
    {      
      ms.addChangeListener(this);
    }
    
  }
  
  /**
   * Das Ergebis der Anfrage wird ermittelt, aufbereitet und versendet
   * 
   * @return true, wenn das Versenden der Daten erfolgreich war
   */
  public boolean send()
  {
    
    System.out.println("SenderZeitGesteuert.send()");
    
    try
    {
      _connection.sendData(_resultData);
      _debug.fine("send : " + _resultData.getData());
    }
    catch (DataNotSubscribedException e)
    {
      try
      {
        _connection.subscribeSender(this, _resultData.getObject(), _dd, SenderRole.source());
        _debug.fine("subscribe!");
      }
      catch (OneSubscriptionPerSendData e1)
      {
        // TODO Auto-generated catch block
        e1.printStackTrace();
      }
      try
      {

        _connection.sendData(_resultData);
        _debug.fine("send : " + _resultData.getData());
      }
      catch (DataNotSubscribedException e1)
      {
        // TODO Auto-generated catch block
        e1.printStackTrace();
      }
      catch (SendSubscriptionNotConfirmed e1)
      {
        // TODO Auto-generated catch block
        e1.printStackTrace();
      }
    }
    catch (SendSubscriptionNotConfirmed e)
    {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    return true;

  }
  

  /**
   * Abstrakte Methode, die implementiert werden muss: bernimmt das tatschliche Senden der Daten
   */
  public void dataRequest(SystemObject object, DataDescription dataDescription, byte state)
  {
    _debug.fine("Sender : Objekt " + object.getPid() + " dd " + dataDescription + " state " + state);
    
  }

  /**
   * Abstrakte Methode, die implementiert werden muss: Zeigt, ob die Sendesteuerung erwnscht ist.
   */
  public boolean isRequestSupported(SystemObject so, DataDescription dd)
  {
    return true;
  }


  /* (non-Javadoc)
   * @see java.util.TimerTask#run()
   */
  public void run()
  {    
    Data data = _meldung.getData().createModifiableCopy();
    
    if (_meldung.getEndlicherAutomat().getMomentZustand().toString().equals("Wiedervorlage")){
      
      TextValue tv = data.getItem("Status").asTextValue();
      tv.setText("in Bearbeitung");      
    
      if (_meldung.setStatus(tv))
        data.getItem("Status").asTextValue().setText(tv.getText());
      else
        _debug.error("Fehler Status: " + _meldung.getPid());
    }
    
    _debug.fine("Moment Zustand: " + _meldung.getEndlicherAutomat().getMomentZustand().toString());
    data.getItem("Status").asTextValue().setText(_meldung.getEndlicherAutomat().getMomentZustand().toString());
    
    _meldung.setData(data);
   
    
    // publiziere
    _debug.fine("Publiziere Daten fr Meldung: " + _resultData.getObject().getPid());
    ResultData res1 = new ResultData(_resultData.getObject(), _dd, System.currentTimeMillis(), data);
    
    Vector<ResultData> sendeDaten = BetriebsMeldungApp.getSendeDaten();
    
    if (!_meldung.getEndlicherAutomat().getMomentZustand().toString().equals("erledigt")){

      synchronized (sendeDaten)
      {
        sendeDaten.add(res1);
        sendeDaten.notify();
      }   
    }
    
    //Sender sender = new Sender(_connection, SenderRole.source());
    //sender.sende(_resultData.getObject(), _dd.getAttributeGroup(), _dd.getAspect(), data);     
      
    DynamicObjectType dot = (DynamicObjectType)_datenmodell.getType("typ.meldung");
    AttributeGroup atg = _datenmodell.getAttributeGroup("atg.meldung");
    Aspect asp = _datenmodell.getAspect("asp.parameterVorgabe");
    _dd = new DataDescription(atg, asp);
    
    
    // persistiere
    _debug.fine("Persistiere Daten fr Meldung: " + _resultData.getObject().getPid());     
    ResultData res2 = new ResultData(_resultData.getObject(), _dd, System.currentTimeMillis(), data);
    
    Vector<ResultData> paramDaten = BetriebsMeldungApp.getParamDaten();

    if (!_meldung.getEndlicherAutomat().getMomentZustand().toString().equals("erledigt")){
    
      synchronized (paramDaten)
      {
        paramDaten.add(res2);
        paramDaten.notify();
      }
      
    }
    
    // Objekt wird entfernt, wenn erledigt
    if (_meldung.getEndlicherAutomat().getMomentZustand().toString().equals("erledigt")){
 
      VerwaltungDynObj vew = new VerwaltungDynObj(_connection, _datenmodell, BetriebsMeldungApp.getKonfigurationsBereich(), dot, BetriebsMeldungApp.getKonfigurationsObjekt(), _dd.getAttributeGroup(), asp);
      vew.setDynamicObject(_connection.getDataModel().getObject(_meldung.getPid()));
      vew.loesche(_meldung.getPid(), "Meldungen");
      
      synchronized (this)
      {
        try
        {
          if (!_geloescht)
          {
            this.wait(10000); // maximal 10 Sekunden auf die Parameter warten, sonst nichts mehr machen
          }
        }
        catch (InterruptedException e1)
        {
          e1.printStackTrace();
        }
      }
      
      SystemObject soMeldung = _connection.getDataModel().getObject(_meldung.getPid());
      if (soMeldung != null && soMeldung.isValid())
      {        
        _debug.error("Konnte Meldung nicht loeschen: " + _meldung.getPid());
        return;
      }
      
      if (_meldung.getEndlicherAutomat().getMomentZustand().toString().equals("erledigt"))
        BetriebsMeldungApp.getMeldungList().remove(_meldung.getPid());
      
      _geloescht = false;
      
      Map<String, Set<String>> mgz = BetriebsMeldungApp.getMeldungsGruppenZuordnung();
      Map<String, Set<String>> meldungsGruppenAenderung = new HashMap<String, Set<String>>();
      
      for (Map.Entry<String, Set<String>> me : mgz.entrySet())
      {
        String key = me.getKey();
        Set<String> list = new TreeSet<String>(me.getValue());
        
        Set<String> delList = new TreeSet<String>();
        
        for (String s : list)
        {
          if (_meldung.getPid().equals(s)){
            //list.remove(s);
        	  delList.add(s);
          }  
        }
        
        list.removeAll(delList);
        
        if (_meldung.getMeldungsGruppen().contains(me.getKey()))
          meldungsGruppenAenderung.put(me.getKey(), list);
      }
   
      for (Map.Entry<String, Set<String>> me : meldungsGruppenAenderung.entrySet())
      {
        
        atg = _datenmodell.getAttributeGroup("atg.meldungsGruppeMeldung");
        asp = _datenmodell.getAspect("asp.zustand");
        DataDescription dd = new DataDescription(atg, asp);
        Data d = _connection.createData(atg);

        String key = me.getKey();
        Set<String> list = me.getValue();

        Array a = d.getItem("MeldungReferenz").asArray();

        a.setLength(list.size());

        int i = 0;
        for (String s : list)
        {
          SystemObject so = _connection.getDataModel().getObject(s);
          a.getItem(i).asReferenceValue().setSystemObject(so);
          i++;
        }

        SystemObject so = _connection.getDataModel().getObject(key);

        res1 = new ResultData(so, dd, System.currentTimeMillis(), d);
        
        synchronized (sendeDaten)
        {
          sendeDaten.add(res1);        
          sendeDaten.notify();
        }        
 

        asp = _datenmodell.getAspect("asp.parameterVorgabe");
        dd = new DataDescription(atg, asp);

        res2 = new ResultData(so, dd, System.currentTimeMillis(), d);

        synchronized (paramDaten)
        {
          paramDaten.add(res2);
          paramDaten.notify();
        }

      }
      
    }
    
    MutableSet ms = BetriebsMeldungApp.getKonfigurationsObjekt().getMutableSet("Meldungen");

    if (ms != null)
    {
      ms.removeChangeListener(this);
    }

  }

  /* (non-Javadoc)
   * @see de.bsvrz.dav.daf.main.config.MutableSetChangeListener#update(de.bsvrz.dav.daf.main.config.MutableSet, de.bsvrz.dav.daf.main.config.SystemObject[], de.bsvrz.dav.daf.main.config.SystemObject[])
   */
  @Override
  public void update(MutableSet set, SystemObject[] addedObjects, SystemObject[] removedObjects)
  {
    if (set.getName().equals("Meldungen"))
    {
      if (addedObjects != null)
      {         
        for (int i = 0; i < addedObjects.length; i++)
        {
//          System.out.println("added - " + addedObjects[i]);
          SystemObject so = addedObjects[i];
        }         
      }

      if (removedObjects != null)
      {
        for (int i = 0; i < removedObjects.length; i++)
        {
//          System.out.println("removed - " + removedObjects[i]);
          SystemObject so = removedObjects[i];          
          if (_meldung.getPid().equals(so.getPid())){
              _geloescht = true;
              synchronized (this)
              {
                this.notify();
              }
          }
        }
      }
    }   
  }

  

}

